#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Created on Tue Mar 05 2024 11:16:29 by codeskyblue
"""

class UiautoException(Exception):
    pass


class DriverException(UiautoException):
    """Base class for all driver-related exceptions."""
    pass

class IOSDriverException(DriverException): ...
class AndroidDriverException(DriverException): ...
class HarmonyDriverException(DriverException): ...
class AppiumDriverException(DriverException): ...


class MethodError(UiautoException):
    pass


class ElementNotFoundError(MethodError): ...
class RequestError(UiautoException): ...


import asyncio
import hashlib
import json
import logging
from pathlib import Path
from typing import Optional

import httpx
import websockets
from fastapi import APIRouter, BackgroundTasks, HTTPException, Request, WebSocket, WebSocketDisconnect
from fastapi.responses import Response, StreamingResponse
from starlette.background import BackgroundTask

logger = logging.getLogger(__name__)
router = APIRouter()
cache_dir = Path("./cache")

# HTTP 转发
@router.api_route("/proxy/http/{target_url:path}", methods=["GET", "POST", "PUT", "DELETE", "PATCH", "OPTIONS"])
async def proxy_http(request: Request, target_url: str):
    logger.info(f"HTTP target_url: {target_url}")

    async with httpx.AsyncClient(timeout=httpx.Timeout(30.0)) as client:
        body = await request.body() if request.method in {"POST", "PUT", "PATCH", "DELETE"} else None
        headers = {k: v for k, v in request.headers.items() if k.lower() not in {"host", "x-target-url"}}
        headers['accept-encoding'] = ''  # disable gzip
        resp = await client.request(
            request.method,
            target_url,
            content=body,
            headers=headers,
        )
        return Response(content=resp.content, status_code=resp.status_code, headers=dict(resp.headers))


@router.get("/")
@router.get("/android/{path:path}")
@router.get("/ios/{path:path}")
@router.get("/demo/{path:path}")
@router.get("/harmony/{path:path}")
async def proxy_html(request: Request):
    target_url = "https://uiauto.dev/"
    cache = HTTPCache(cache_dir, target_url, key='homepage')
    response = await cache.proxy_request(request, update_cache=True)
    return response
    # update

@router.get("/assets/{path:path}")
@router.get('/favicon.ico')
async def proxy_assets(request: Request, path: str = ""):
    target_url = f"https://uiauto.dev{request.url.path}"
    cache = HTTPCache(cache_dir, target_url)
    return await cache.proxy_request(request)


class HTTPCache:
    def __init__(self, cache_dir: Path, target_url: str, key: Optional[str] = None):
        self.cache_dir = cache_dir
        self.target_url = target_url
        self.key = key or hashlib.md5(target_url.encode()).hexdigest()
        self.file_body = self.cache_dir / 'http' / (self.key + ".body")
        self.file_headers = self.file_body.with_suffix(".headers")

    async def proxy_request(self, request: Request, update_cache: bool = False):
        response = await self.get_cached_response(request)
        if not response:
            response = await self.proxy_and_save_response(request)
            return response
        if update_cache:
            # async update cache in background
            asyncio.create_task(self.update_cache(request))
        return response

    async def get_cached_response(self, request: Request):
        if request.method == 'GET' and self.file_body.exists():
            logger.info(f"Cache hit: {self.file_body}")
            headers = {}
            if self.file_headers.exists():
                with self.file_headers.open('rb') as f:
                    headers = json.load(f)
            body_fd = self.file_body.open("rb")
            return StreamingResponse(
                content=body_fd,
                status_code=200,
                headers=headers,
                background=BackgroundTask(body_fd.close)
            )
        return None

    async def update_cache(self, request: Request):
        try:
            await self.proxy_and_save_response(request)
        except Exception as e:
            logger.error("Update cache failed")

    async def proxy_and_save_response(self, request: Request) -> Response:
        logger.debug(f"Proxying request... {request.url.path}")
        response = await proxy_http(request, self.target_url)
        # save response to cache
        if request.method == "GET" and response.status_code == 200 and self.cache_dir.exists():
            self.file_body.parent.mkdir(parents=True, exist_ok=True)
            with self.file_body.open("wb") as f:
                f.write(response.body)
            with self.file_headers.open("w", encoding="utf-8") as f:
                headers = response.headers
                headers['cache-status'] = 'HIT'
                json.dump(dict(headers), f, indent=2, ensure_ascii=False)
        return response


# WebSocket 转发
@router.websocket("/proxy/ws/{target_url:path}")
async def proxy_ws(websocket: WebSocket, target_url: str):
    await websocket.accept()
    logger.info(f"WebSocket target_url: {target_url}")

    try:
        async with websockets.connect(target_url) as target_ws:
            async def from_client():
                while True:
                    msg = await websocket.receive_text()
                    await target_ws.send(msg)

            async def from_server():
                while True:
                    msg = await target_ws.recv()
                    if isinstance(msg, bytes):
                        await websocket.send_bytes(msg)
                    elif isinstance(msg, str):
                        await websocket.send_text(msg)
                    else:
                        raise RuntimeError("Unknown message type", msg)

            await asyncio.gather(from_client(), from_server())

    except WebSocketDisconnect:
        pass
    except Exception as e:
        logger.error(f"WS Error: {e}")
        await websocket.close()

# ref: https://stackoverflow.com/questions/74555102/how-to-forward-fastapi-requests-to-another-server
def make_reverse_proxy(base_url: str, strip_prefix: str = ""):
    async def _reverse_proxy(request: Request):
        client = httpx.AsyncClient(base_url=base_url)
        client.timeout = httpx.Timeout(30.0, read=300.0)
        path = request.url.path
        if strip_prefix and path.startswith(strip_prefix):
            path = path[len(strip_prefix):]
        target_url = httpx.URL(
            path=path, query=request.url.query.encode("utf-8")
        )
        exclude_headers = [b"host", b"connection", b"accept-encoding"]
        headers = [(k, v) for k, v in request.headers.raw if k not in exclude_headers]
        headers.append((b'accept-encoding', b''))
        
        req = client.build_request(
            request.method, target_url, headers=headers, content=request.stream()
        )
        r = await client.send(req, stream=True)#, follow_redirects=True)
        
        response_headers = {
            k: v for k, v in r.headers.items()
            if k.lower() not in {"transfer-encoding", "connection", "content-length"}
        }
        async def gen_content():
            async for chunk in r.aiter_bytes(chunk_size=40960):
                yield chunk
        
        async def aclose():
            await client.aclose()
        
        return StreamingResponse(
            content=gen_content(),
            status_code=r.status_code,
            headers=response_headers,
            background=BackgroundTask(aclose),
        )

    return _reverse_proxy



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Created on Fri Mar 01 2024 14:19:29 by codeskyblue
"""

import logging
import re
import time
from typing import Iterator, List, Optional, Tuple

import adbutils
from PIL import Image

from uiautodev.command_types import CurrentAppResponse
from uiautodev.driver.android.common import parse_xml
from uiautodev.driver.base_driver import BaseDriver
from uiautodev.exceptions import AndroidDriverException
from uiautodev.model import AppInfo, Node, Rect, ShellResponse, WindowSize

logger = logging.getLogger(__name__)

class ADBAndroidDriver(BaseDriver):
    def __init__(self, serial: str):
        super().__init__(serial)
        self.adb_device = adbutils.device(serial)

    def get_current_activity(self) -> str:
        ret = self.adb_device.shell2(["dumpsys", "activity", "activities"], rstrip=True, timeout=5)
        # 使用正则查找包含前台 activity 的行
        match = re.search(r"mResumedActivity:.*? ([\w\.]+\/[\w\.]+)", ret.output)
        if match:
            return match.group(1)  # 返回包名/类名，例如 com.example/.MainActivity
        else:
            return ""
    
    def screenshot(self, id: int) -> Image.Image:
        if id > 0:
            raise AndroidDriverException("multi-display is not supported yet for uiautomator2")
        return self.adb_device.screenshot(display_id=id)

    def shell(self, command: str) -> ShellResponse:
        try:
            ret = self.adb_device.shell2(command, rstrip=True, timeout=20)
            if ret.returncode == 0:
                return ShellResponse(output=ret.output, error=None)
            else:
                return ShellResponse(
                    output="", error=f"exit:{ret.returncode}, output:{ret.output}"
                )
        except Exception as e:
            return ShellResponse(output="", error=f"adb error: {str(e)}")

    def dump_hierarchy(self, display_id: Optional[int] = 0) -> Tuple[str, Node]:
        """returns xml string and hierarchy object"""
        start = time.time()
        try:
            xml_data = self._dump_hierarchy_raw()
            logger.debug("dump_hierarchy cost: %s", time.time() - start)
        except Exception as e:
            raise AndroidDriverException(f"Failed to dump hierarchy: {str(e)}")

        wsize = self.adb_device.window_size()
        logger.debug("window size: %s", wsize)
        return xml_data, parse_xml(
            xml_data, WindowSize(width=wsize[0], height=wsize[1]), display_id
        )

    def _dump_hierarchy_raw(self) -> str:
        """
        uiautomator2 server is conflict with "uiautomator dump" command.

        uiautomator dump errors:
        - ERROR: could not get idle state.
        """
        try:
            return self.adb_device.dump_hierarchy()
        except adbutils.AdbError as e:
            if "Killed" in str(e):
                self.kill_app_process()
            return self.adb_device.dump_hierarchy()
    
    def kill_app_process(self):
        logger.debug("Killing app_process")
        pids = []
        for line in self.adb_device.shell("ps -A || ps").splitlines():
            if "app_process" in line:
                fields = line.split()
                if len(fields) >= 2:
                    pids.append(int(fields[1]))
                    logger.debug(f"App process PID: {fields[1]}")
        for pid in set(pids):
            self.adb_device.shell(f"kill {pid}")

    def tap(self, x: int, y: int):
        self.adb_device.click(x, y)

    def window_size(self) -> Tuple[int, int]:
        w, h = self.adb_device.window_size()
        return (w, h)

    def app_install(self, app_path: str):
        self.adb_device.install(app_path)

    def app_current(self) -> CurrentAppResponse:
        info = self.adb_device.app_current()
        return CurrentAppResponse(
            package=info.package, activity=info.activity, pid=info.pid
        )

    def app_launch(self, package: str):
        if self.adb_device.package_info(package) is None:
            raise AndroidDriverException(f"App not installed: {package}")
        self.adb_device.app_start(package)
    
    def app_terminate(self, package: str):
        self.adb_device.app_stop(package)

    def home(self):
        self.adb_device.keyevent("HOME")
    
    def wake_up(self):
        self.adb_device.keyevent("WAKEUP")
    
    def back(self):
        self.adb_device.keyevent("BACK")
    
    def app_switch(self):
        self.adb_device.keyevent("APP_SWITCH")
    
    def volume_up(self):
        self.adb_device.keyevent("VOLUME_UP")
    
    def volume_down(self):
        self.adb_device.keyevent("VOLUME_DOWN")
    
    def volume_mute(self):
        self.adb_device.keyevent("VOLUME_MUTE")

    def get_app_version(self, package_name: str) -> dict:
        """
        Get the version information of an app, including mainVersion and subVersion.

        Args:
            package_name (str): The package name of the app.

        Returns:
            dict: A dictionary containing mainVersion and subVersion.
        """
        output = self.adb_device.shell(["dumpsys", "package", package_name])

        # versionName
        m = re.search(r"versionName=(?P<name>[^\s]+)", output)
        version_name = m.group("name") if m else ""
        if version_name == "null":  # Java dumps "null" for null values
            version_name = None

        # versionCode
        m = re.search(r"versionCode=(?P<code>\d+)", output)
        version_code = m.group("code") if m else ""
        version_code = int(version_code) if version_code.isdigit() else None

        return {
            "versionName": version_name,
            "versionCode": version_code
        }

    def app_list(self) -> List[AppInfo]:
        results = []
        output = self.adb_device.shell(["pm", "list", "packages", '-3'])
        for m in re.finditer(r"^package:([^\s]+)\r?$", output, re.M):
            packageName = m.group(1)
            # get version
            version_info = self.get_app_version(packageName)
            app_info = AppInfo(
                packageName=packageName,
                versionName=version_info.get("versionName"),
                versionCode=version_info.get("versionCode")
            )
            results.append(app_info)
        return results

    def open_app_file(self, package: str) -> Iterator[bytes]:
        line = self.adb_device.shell(f"pm path {package}")
        assert isinstance(line, str)
        if not line.startswith("package:"):
            raise AndroidDriverException(f"Failed to get package path: {line}")
        remote_path = line.split(':', 1)[1]
        yield from self.adb_device.sync.iter_content(remote_path)
    
    def send_keys(self, text: str):
        self.adb_device.send_keys(text)
    
    def clear_text(self):
        for _ in range(3):
            self.adb_device.shell2("input keyevent DEL --longpress")



import re
from functools import partial
from typing import List, Optional, Tuple
from xml.etree import ElementTree

from uiautodev.exceptions import AndroidDriverException, RequestError
from uiautodev.model import AppInfo, Node, Rect, ShellResponse, WindowSize


def parse_xml(xml_data: str, wsize: WindowSize, display_id: Optional[int] = None) -> Node:
    root = ElementTree.fromstring(xml_data)
    node = parse_xml_element(root, wsize, display_id)
    if node is None:
        raise AndroidDriverException("Failed to parse xml")
    return node


def parse_xml_element(element, wsize: WindowSize, display_id: Optional[int], indexes: List[int] = [0]) -> Optional[Node]:
    """
    Recursively parse an XML element into a dictionary format.
    """
    name = element.tag
    if name == "node":
        name = element.attrib.get("class", "node")
    if display_id is not None:
        elem_display_id = int(element.attrib.get("display-id", display_id))
        if elem_display_id != display_id:
            return

    bounds = None
    rect = None
    # eg: bounds="[883,2222][1008,2265]"
    if "bounds" in element.attrib:
        bounds = element.attrib["bounds"]
        bounds = list(map(int, re.findall(r"\d+", bounds)))
        assert len(bounds) == 4
        rect = Rect(x=bounds[0], y=bounds[1], width=bounds[2] - bounds[0], height=bounds[3] - bounds[1])
        bounds = (
            bounds[0] / wsize.width,
            bounds[1] / wsize.height,
            bounds[2] / wsize.width,
            bounds[3] / wsize.height,
        )
        bounds = map(partial(round, ndigits=4), bounds)
        
    elem = Node(
        key="-".join(map(str, indexes)),
        name=name,
        bounds=bounds,
        rect=rect,
        properties={key: element.attrib[key] for key in element.attrib},
        children=[],
    )

    # Construct xpath for children
    for index, child in enumerate(element):
        child_node = parse_xml_element(child, wsize, display_id, indexes + [index])
        if child_node:
            elem.children.append(child_node)

    return elem



# prefix for /api/android/{serial}/shell

import logging
from typing import Dict, Optional

from fastapi import APIRouter, Request, Response
from pydantic import BaseModel

from uiautodev.driver.android import ADBAndroidDriver, U2AndroidDriver
from uiautodev.model import ShellResponse

logger = logging.getLogger(__name__)

router = APIRouter()


class AndroidShellPayload(BaseModel):
    command: str
    
@router.post("/{serial}/shell")
def shell(serial: str, payload: AndroidShellPayload) -> ShellResponse:
    """Run a shell command on an Android device"""
    try:
        driver = ADBAndroidDriver(serial)
        return driver.shell(payload.command)
    except NotImplementedError as e:
        return Response(content="shell not implemented", media_type="text/plain", status_code=501)
    except Exception as e:
        logger.exception("shell failed")
        return ShellResponse(output="", error=str(e))


@router.get("/{serial}/current_activity")
async def get_current_activity(serial: str) -> Response:
    """Get the current activity of the Android device"""
    try:
        driver = ADBAndroidDriver(serial)
        activity = driver.get_current_activity()
        return Response(content=activity, media_type="text/plain")
    except Exception as e:
        logger.exception("get_current_activity failed")
        return Response(content="", media_type="text/plain")


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Created on Sun Feb 18 2024 13:48:55 by codeskyblue"""

import logging
import os
import platform
import signal
from pathlib import Path
from typing import Dict, List

import adbutils
import httpx
import uvicorn
from fastapi import FastAPI, File, Request, Response, UploadFile, WebSocket
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import FileResponse, JSONResponse, RedirectResponse
from pydantic import BaseModel
from starlette.websockets import WebSocketDisconnect

from uiautodev import __version__
from uiautodev.common import convert_bytes_to_image, get_webpage_url, ocr_image
from uiautodev.driver.android import ADBAndroidDriver, U2AndroidDriver
from uiautodev.model import Node
from uiautodev.provider import AndroidProvider, HarmonyProvider, IOSProvider, MockProvider
from uiautodev.remote.scrcpy import ScrcpyServer
from uiautodev.router.android import router as android_device_router
from uiautodev.router.device import make_router
from uiautodev.router.proxy import make_reverse_proxy
from uiautodev.router.proxy import router as proxy_router
from uiautodev.router.xml import router as xml_router
from uiautodev.utils.envutils import Environment

logger = logging.getLogger(__name__)

app = FastAPI()

app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["GET", "POST"],
    allow_headers=["*"],
)

android_default_driver = U2AndroidDriver
if os.getenv("UIAUTODEV_USE_ADB_DRIVER") in ("1", "true", "True"):
    android_default_driver = ADBAndroidDriver

android_router = make_router(AndroidProvider(driver_class=android_default_driver))
android_adb_router = make_router(AndroidProvider(driver_class=ADBAndroidDriver))
ios_router = make_router(IOSProvider())
harmony_router = make_router(HarmonyProvider())
mock_router = make_router(MockProvider())

app.include_router(mock_router, prefix="/api/mock", tags=["mock"])

if Environment.UIAUTODEV_MOCK:
    app.include_router(mock_router, prefix="/api/android", tags=["mock"])
    app.include_router(mock_router, prefix="/api/ios", tags=["mock"])
    app.include_router(mock_router, prefix="/api/harmony", tags=["mock"])
else:
    app.include_router(android_router, prefix="/api/android", tags=["android"])
    app.include_router(android_adb_router, prefix="/api/android_adb", tags=["android_adb"])
    app.include_router(ios_router, prefix="/api/ios", tags=["ios"])
    app.include_router(harmony_router, prefix="/api/harmony", tags=["harmony"])

app.include_router(xml_router, prefix="/api/xml", tags=["xml"])
app.include_router(android_device_router, prefix="/api/android", tags=["android"])
app.include_router(proxy_router, tags=["proxy"])


@app.get("/api/{platform}/features")
def get_features(platform: str) -> Dict[str, bool]:
    """Get features supported by the specified platform"""
    features = {}
    # 获取所有带有指定平台tag的路由
    from starlette.routing import Route

    for route in app.routes:
        _route: Route = route  # type: ignore
        if hasattr(_route, "tags") and platform in _route.tags:
            if _route.path.startswith(f"/api/{platform}/{{serial}}/"):
                # 提取特性名称
                parts = _route.path.split("/")
                feature_name = parts[-1]
                if not feature_name.startswith("{"):
                    features[feature_name] = True
    return features


class InfoResponse(BaseModel):
    version: str
    description: str
    platform: str
    code_language: str
    cwd: str
    drivers: List[str]


@app.get("/api/info")
def info() -> InfoResponse:
    """Information about the application"""
    return InfoResponse(
        version=__version__,
        description="client for https://uiauto.dev",
        platform=platform.system(),  # Linux | Darwin | Windows
        code_language="Python",
        cwd=os.getcwd(),
        drivers=["android", "ios", "harmony"],
    )


@app.post("/api/ocr_image")
async def _ocr_image(file: UploadFile = File(...)) -> List[Node]:
    """OCR an image"""
    image_data = await file.read()
    image = convert_bytes_to_image(image_data)
    return ocr_image(image)


@app.get("/shutdown")
def shutdown() -> str:
    """Shutdown the server"""
    os.kill(os.getpid(), signal.SIGINT)
    return "Server shutting down..."


@app.get("/demo")
def demo():
    """Demo endpoint"""
    static_dir = Path(__file__).parent / "static"
    print(static_dir / "demo.html")
    return FileResponse(static_dir / "demo.html")


@app.get("/redirect")
def index_redirect():
    """redirect to official homepage"""
    url = get_webpage_url()
    logger.debug("redirect to %s", url)
    return RedirectResponse(url)


@app.get("/api/auth/me")
def mock_auth_me():
    # 401 {"detail":"Authentication required"}
    return JSONResponse(status_code=401, content={"detail": "Authentication required"})

@app.websocket("/ws/android/scrcpy/{serial}")
async def handle_android_ws(websocket: WebSocket, serial: str):
    """
    Args:
        serial: device serial
        websocket: WebSocket
    """
    await websocket.accept()

    try:
        logger.info(f"WebSocket serial: {serial}")
        device = adbutils.device(serial)
        server = ScrcpyServer(device)
        await server.handle_unified_websocket(websocket, serial)
    except WebSocketDisconnect:
        logger.info(f"WebSocket disconnected by client.")
    except Exception as e:
        logger.exception(f"WebSocket error for serial={serial}: {e}")
        await websocket.close(code=1000, reason=str(e))
    finally:
        logger.info(f"WebSocket closed for serial={serial}")


def get_harmony_mjpeg_server(serial: str):
    from hypium import UiDriver

    from uiautodev.remote.harmony_mjpeg import HarmonyMjpegServer

    driver = UiDriver.connect(device_sn=serial)
    logger.info("create harmony mjpeg server for %s", serial)
    logger.info(f"device wake_up_display: {driver.wake_up_display()}")
    return HarmonyMjpegServer(driver)


@app.websocket("/ws/harmony/mjpeg/{serial}")
async def unified_harmony_ws(websocket: WebSocket, serial: str):
    """
    Args:
        serial: device serial
        websocket: WebSocket
    """
    await websocket.accept()

    try:
        logger.info(f"WebSocket serial: {serial}")

        # 获取 HarmonyScrcpyServer 实例
        server = get_harmony_mjpeg_server(serial)
        server.start()
        await server.handle_ws(websocket)
    except ImportError as e:
        logger.error(f"missing library for harmony: {e}")
        await websocket.close(
            code=1000, reason='missing library, fix by "pip install uiautodev[harmony]"'
        )
    except WebSocketDisconnect:
        logger.info(f"WebSocket disconnected by client.")
    except Exception as e:
        logger.exception(f"WebSocket error for serial={serial}: {e}")
        await websocket.close(code=1000, reason=str(e))
    finally:
        logger.info(f"WebSocket closed for serial={serial}")


if __name__ == "__main__":
    uvicorn.run("uiautodev.app:app", port=4000, reload=True, use_colors=True)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Created on Fri Mar 01 2024 14:00:10 by codeskyblue
"""

import io
import logging
from typing import Any, Dict, List, Optional

from fastapi import APIRouter, Query, Request, Response
from fastapi.responses import StreamingResponse
from pydantic import BaseModel

from uiautodev import command_proxy
from uiautodev.command_types import Command, CurrentAppResponse, InstallAppRequest, InstallAppResponse, TapRequest
from uiautodev.model import DeviceInfo, Node, ShellResponse
from uiautodev.provider import BaseProvider

logger = logging.getLogger(__name__)


def make_router(provider: BaseProvider) -> APIRouter:
    router = APIRouter()

    @router.get("/list")
    def _list() -> List[DeviceInfo]:
        """List devices"""
        try:
            return provider.list_devices()
        except NotImplementedError as e:
            return Response(content="list_devices not implemented", media_type="text/plain", status_code=501)
        except Exception as e:
            logger.exception("list_devices failed")
            return Response(content=str(e), media_type="text/plain", status_code=500)

    @router.get(
        "/{serial}/screenshot/{id}",
        responses={200: {"content": {"image/jpeg": {}}}},
        response_class=Response,
    )
    def _screenshot(serial: str, id: int) -> Response:
        """Take a screenshot of device"""
        try:
            driver = provider.get_device_driver(serial)
            pil_img = driver.screenshot(id).convert("RGB")
            buf = io.BytesIO()
            pil_img.save(buf, format="JPEG")
            image_bytes = buf.getvalue()
            return Response(content=image_bytes, media_type="image/jpeg")
        except Exception as e:
            logger.exception("screenshot failed")
            return Response(content=str(e), media_type="text/plain", status_code=500)

    @router.get("/{serial}/hierarchy")
    def dump_hierarchy(serial: str, format: str = "json"):
        """Dump the view hierarchy of an Android device"""
        try:
            driver = provider.get_device_driver(serial)
            xml_data, hierarchy = driver.dump_hierarchy()
            if format == "xml":
                return Response(content=xml_data, media_type="text/xml")
            elif format == "json":
                return hierarchy
            else:
                return Response(content=f"Invalid format: {format}", media_type="text/plain", status_code=400)
        except Exception as e:
            #logger.exception("dump_hierarchy failed")
            logger.error(f"Error dumping hierarchy: {str(e)}")
            return Response(content=str(e), media_type="text/plain", status_code=500)
    
    @router.post('/{serial}/command/tap')
    def command_tap(serial: str, params: TapRequest):
        """Run a command on the device"""
        driver = provider.get_device_driver(serial)
        command_proxy.tap(driver, params)
        return {"status": "ok"}
    
    @router.post('/{serial}/command/installApp')
    def install_app(serial: str, params: InstallAppRequest) -> InstallAppResponse:
        """Install app"""
        driver = provider.get_device_driver(serial)
        return command_proxy.app_install(driver, params)

    @router.get('/{serial}/command/currentApp')
    def current_app(serial: str) -> CurrentAppResponse:
        """Get current app"""
        driver = provider.get_device_driver(serial)
        return command_proxy.app_current(driver)

    @router.post('/{serial}/command/{command}')
    def _command_proxy_other(serial: str, command: Command, params: Dict[str, Any] = None):
        """Run a command on the device"""
        driver = provider.get_device_driver(serial)
        response = command_proxy.send_command(driver, command, params)
        return response
    
    @router.get('/{serial}/backupApp')
    def _backup_app(serial: str, packageName: str):
        """Backup app
        
        Added in 0.5.0
        """
        driver = provider.get_device_driver(serial)
        file_name = f"{packageName}.apk"
        headers = {
            'Content-Disposition': f'attachment; filename="{file_name}"'
        }
        return StreamingResponse(driver.open_app_file(packageName), headers=headers)
        


    return router
